import json
from typing import Dict, List, Optional, Union

from django.db.models import (
    DateField,
    Field,
    ForeignKey,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel,
    Model,
    OneToOneField,
    OneToOneRel,
    TimeField,
)
from django.db.models.fields import AutoFieldMixin
from forestadmin.agent_toolkit.forest_logger import ForestLogger
from forestadmin.datasource_django.utils.type_converter import FilterOperator, TypeConverter
from forestadmin.datasource_toolkit.interfaces.fields import (
    Column,
    FieldType,
    ManyToMany,
    ManyToOne,
    OneToMany,
    OneToOne,
    Operator,
    Validation,
)
from forestadmin.datasource_toolkit.interfaces.models.collections import CollectionSchema

try:
    # GeneratedField is available since django 5
    from django.db.models import GeneratedField
except ImportError:
    GeneratedField = None


class FieldFactory:
    @staticmethod
    def _build_enum_values(field: Field) -> Optional[List[str]]:
        if field.choices:
            return [str(c[0]) for c in field.choices]  # type: ignore
        return None

    @staticmethod
    def _build_is_read_only(field: Field) -> bool:
        if field.primary_key:
            return isinstance(field, AutoFieldMixin)
        elif isinstance(field, DateField) or isinstance(field, TimeField):
            return field.auto_now is True or field.auto_now_add is True
        if GeneratedField is not None and isinstance(field, GeneratedField):
            return True
        return not field.editable

    @classmethod
    def _build_validations(cls, field: Field) -> List[Validation]:
        validations: List[Validation] = []
        if not (field.null or field.blank) and not cls._build_is_read_only(field) and not field.has_default():
            validations.append(
                {
                    "operator": Operator.PRESENT,
                }
            )
        if field.max_length:
            validations.append({"operator": Operator.SHORTER_THAN, "value": field.max_length})

        return validations

    @classmethod
    def build(cls, field: Field, current_model: Model) -> Column:
        column_type = TypeConverter.convert(field)  # type: ignore

        default_value = field.default
        try:
            # field.get_default() evaluate a function if default is a function
            # we prefer don't have a default value rather than a false one
            json.dumps(default_value)
        except TypeError:  # not JSON Serializable
            default_value = None

        # when model inheritance (https://docs.djangoproject.com/en/5.0/topics/db/models/#multi-table-inheritance),
        # you inherit from parent model attributes, and also his pk ;
        # but the pk of the parent model is not the pk of the current model!
        is_primary_key = field.primary_key and current_model == field.model

        return {
            "column_type": column_type,
            "is_primary_key": is_primary_key,  # type: ignore
            "is_read_only": cls._build_is_read_only(field),
            "default_value": default_value,
            "is_sortable": True,
            "validations": cls._build_validations(field),
            "filter_operators": FilterOperator.get_for_type(column_type),
            "enum_values": cls._build_enum_values(field),
            "type": FieldType.COLUMN,
        }


class DjangoCollectionFactory:
    @staticmethod
    def _build_one_to_many(relation: ManyToOneRel) -> Optional[OneToMany]:
        return {
            "foreign_collection": relation.target_field.model._meta.db_table,
            "origin_key": relation.field.attname,
            "origin_key_target": relation.field.target_field.attname,
            "type": FieldType.ONE_TO_MANY,
        }

    @staticmethod
    def _build_many_to_one(relation: Union[OneToOneField, ForeignKey, ManyToOneRel]) -> Optional[ManyToOne]:
        if isinstance(relation, ManyToOneRel):
            foreign_key = relation.field.attname
        elif isinstance(relation, ForeignKey) or isinstance(relation, OneToOneField):
            foreign_key = relation.attname
        return {
            "foreign_collection": relation.target_field.model._meta.db_table,
            "foreign_key": foreign_key,
            "foreign_key_target": relation.target_field.attname,
            "type": FieldType.MANY_TO_ONE,
        }

    @staticmethod
    def _build_one_to_one(relation: OneToOneRel) -> OneToOne:
        return {
            "foreign_collection": relation.target_field.model._meta.db_table,
            "origin_key": relation.field.attname,
            "origin_key_target": relation.field.target_field.attname,
            "type": FieldType.ONE_TO_ONE,
        }

    @staticmethod
    def _build_many_to_many(relation: Union[ManyToManyField, ManyToManyRel]) -> ManyToMany:
        kwargs: Dict[str, str] = {"foreign_collection": relation.target_field.model._meta.db_table}

        if isinstance(relation, ManyToManyField):
            kwargs["through_collection"] = relation.remote_field.through._meta.db_table

            kwargs["origin_key"] = relation.m2m_column_name()
            kwargs["foreign_key"] = relation.m2m_reverse_name()

        elif isinstance(relation, ManyToManyRel):
            kwargs["through_collection"] = relation.through._meta.db_table

            kwargs["origin_key"] = relation.field.m2m_reverse_name()
            kwargs["foreign_key"] = relation.field.m2m_column_name()

        kwargs["origin_key_target"] = relation.model._meta.pk.attname
        kwargs["foreign_key_target"] = relation.target_field.model._meta.pk.attname

        return ManyToMany(type=FieldType.MANY_TO_MANY, foreign_relation=None, **kwargs)

    @staticmethod
    def build(model: Model) -> CollectionSchema:
        fields = {}
        for field in model._meta.get_fields(include_hidden=True):
            if not field.is_relation:
                fields[field.name] = FieldFactory.build(field, model)
            else:
                if generic_foreign_key(field) or is_generic_rel(field) or is_generic_relation(field):
                    ForestLogger.log(
                        "warning",
                        f"Ignoring {model._meta.db_table}.{field.name} because polymorphic relation is not supported.",
                    )
                    continue

                # get_fields with include hidden doesn't include autogenerated foreign key fields (ending with _id)
                if isinstance(field, ForeignKey):
                    fields[field.attname] = FieldFactory.build(field, model)

                if field.one_to_one is True:
                    if isinstance(field, OneToOneField):
                        fields[field.name] = DjangoCollectionFactory._build_many_to_one(field)
                    else:
                        fields[field.name] = DjangoCollectionFactory._build_one_to_one(field)

                elif field.one_to_many is True:
                    fields[f"{field.name}_{field.remote_field.name}"] = DjangoCollectionFactory._build_one_to_many(
                        field
                    )

                elif field.many_to_one is True:
                    fields[field.name] = DjangoCollectionFactory._build_many_to_one(field)

                elif field.many_to_many is True:
                    fields[field.name] = DjangoCollectionFactory._build_many_to_many(field)

        return {"actions": {}, "fields": fields, "searchable": False, "segments": []}


def generic_foreign_key(field):
    # when imported at top level (before app.Ready):
    # it raise django.core.exceptions.AppRegistryNotReady: Apps aren't loaded yet.
    from django.contrib.contenttypes.fields import GenericForeignKey

    return isinstance(field, GenericForeignKey)


def is_generic_relation(field):
    # when imported at top level (before app.Ready):
    # it raise django.core.exceptions.AppRegistryNotReady: Apps aren't loaded yet.
    from django.contrib.contenttypes.fields import GenericRelation

    return isinstance(field, GenericRelation)


def is_generic_rel(field):
    # when imported at top level (before app.Ready):
    # it raise django.core.exceptions.AppRegistryNotReady: Apps aren't loaded yet.
    from django.contrib.contenttypes.fields import GenericRel

    return isinstance(field, GenericRel)



from django.apps import apps
from forestadmin.datasource_django.collection import DjangoCollection
from forestadmin.datasource_django.interface import BaseDjangoDatasource


class DjangoDatasource(BaseDjangoDatasource):
    def __init__(self) -> None:
        super().__init__()
        self._create_collections()

    def _create_collections(self):
        models = apps.get_models(include_auto_created=True)
        for model in models:
            if model._meta.proxy is False:
                collection = DjangoCollection(self, model)
                self.add_collection(collection)



from typing import List, Optional

from asgiref.sync import sync_to_async
from django.db import connections
from django.db.models import Model
from forestadmin.agent_toolkit.utils.context import User
from forestadmin.datasource_django.interface import BaseDjangoCollection
from forestadmin.datasource_django.utils.model_introspection import DjangoCollectionFactory
from forestadmin.datasource_django.utils.native_driver_wrapper import NativeDriverWrapper, get_db_for_native_driver
from forestadmin.datasource_django.utils.query_factory import DjangoQueryBuilder
from forestadmin.datasource_django.utils.record_serializer import instance_to_record_data
from forestadmin.datasource_toolkit.datasources import Datasource
from forestadmin.datasource_toolkit.interfaces.fields import is_column
from forestadmin.datasource_toolkit.interfaces.query.aggregation import AggregateResult, Aggregation
from forestadmin.datasource_toolkit.interfaces.query.filter.paginated import PaginatedFilter
from forestadmin.datasource_toolkit.interfaces.query.filter.unpaginated import Filter
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias


class DjangoCollection(BaseDjangoCollection):
    def __init__(self, datasource: Datasource, model: Model):
        super().__init__(model._meta.db_table, datasource)
        self._model = model
        schema = DjangoCollectionFactory.build(model)
        self.add_fields(schema["fields"])

    @property
    def model(self) -> Model:
        return self._model

    async def list(self, caller: User, filter_: PaginatedFilter, projection: Projection) -> List[RecordsDataAlias]:
        return [
            await sync_to_async(instance_to_record_data)(item, projection)
            for item in await DjangoQueryBuilder.mk_list(self, filter_, projection)
        ]

    async def aggregate(
        self, caller: User, filter_: Optional[Filter], aggregation: Aggregation, limit: Optional[int] = None
    ) -> List[AggregateResult]:
        return await DjangoQueryBuilder.mk_aggregate(self, filter_, aggregation, limit)

    async def create(self, caller: User, data: List[RecordsDataAlias]) -> List[RecordsDataAlias]:
        instances = await DjangoQueryBuilder.mk_create(self, data)
        projection = Projection(*[k for k in self.schema["fields"].keys() if is_column(self.schema["fields"][k])])
        return [await sync_to_async(instance_to_record_data)(item, projection) for item in instances]

    async def update(self, caller: User, filter_: Optional[Filter], patch: RecordsDataAlias) -> None:
        await DjangoQueryBuilder.mk_update(self, filter_, patch)

    async def delete(self, caller: User, filter_: Optional[Filter]) -> None:
        await DjangoQueryBuilder.mk_delete(self, filter_)

    def get_native_driver(self) -> NativeDriverWrapper:
        db_name = get_db_for_native_driver(self.model)
        return NativeDriverWrapper(connections[db_name])



from decimal import Decimal
from typing import Any

from django.db.models import Model
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias


def instance_to_record_data(instance: Model, projection: Projection) -> RecordsDataAlias:
    record_data = {}
    for field_name in projection.columns:
        record_data[field_name] = serialize_value(getattr(instance, field_name))

    for relation_name, subfields in projection.relations.items():
        relation = getattr(instance, relation_name, None)
        if relation:
            record_data[relation_name] = instance_to_record_data(relation, subfields)
        else:
            record_data[relation_name] = None

    return record_data


def serialize_value(value: Any):
    if isinstance(value, Decimal):
        return float(value)
    return value
